"""
This module provides controls for projects information.

neoteroi.projects

MIT License
Copyright (c) 2022 to present, Roberto Prevato
"""
from markdown import Extension

from .gantt import register_extension


class ProjectsExtension(Extension):

    config = {
        "priority": [12, "The priority to be configured for the extension."],
    }

    def extendMarkdown(self, md):
        md.registerExtension(self)
        register_extension(md, self.getConfig("priority"))



import re
from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Tuple

from .. import extract_props
from . import Matrix, Table

SPAN_RE = re.compile(r"@span=?([\d]+)?:?([\d]+)?")


@dataclass
class Cell:
    text: str
    skip: bool = False
    col_span: int = 1
    row_span: int = 1
    props: Optional[Dict[str, str]] = None

    @property
    def html_class(self) -> Optional[str]:
        return self.props.get("class") if self.props else None


def _iter_coords(
    x: int, y: int, colspan: int, rowspan: int
) -> Iterable[Tuple[int, int]]:

    for x_increment in range(colspan):
        for y_increment in range(rowspan):
            yield (x + x_increment, y + y_increment)


def _get_auto_cols_span(values: Tuple[str, ...], start_index: int = 0) -> int:
    span = 1

    for i in range(start_index, len(values)):
        if not values[i].strip():
            span += 1
        else:
            break
    return span


def _get_auto_rows_span(
    rows: List[Tuple[str, ...]], start_index: int, cols_slice: slice
) -> int:
    span = 1

    for i in range(start_index, len(rows)):
        values_to_check = rows[i][cols_slice]
        if all(not value.strip() for value in values_to_check):
            span += 1
        else:
            break
    return span


def get_matrix(table: Table) -> Matrix:
    matrix = Matrix(len(table.records[0]), len(table.records) + 1)
    rows = list(table.records)
    rows.insert(0, table.headers)

    for row_index, row in enumerate(rows):
        for column_index, value in enumerate(row):
            existing_cell = matrix[column_index, row_index]

            if existing_cell is not None:
                # set by previous span
                continue

            match = SPAN_RE.search(value)
            if match:
                raw_cols_span = match.group(1)
                raw_rows_span = match.group(2)

                if raw_cols_span is None and raw_rows_span is None:
                    # Automatic mode: increase the span until empty cells are found
                    # columns span takes precedence over rows span
                    cols_span = _get_auto_cols_span(row, column_index + 1)
                    rows_span = _get_auto_rows_span(
                        rows,
                        row_index + 1,
                        slice(column_index, column_index + cols_span),
                    )
                else:
                    cols_span = max(int(raw_cols_span or 1), 1)
                    rows_span = max(int(raw_rows_span or 1), 1)

                if cols_span > 0 or rows_span > 0:
                    for coords in _iter_coords(
                        column_index, row_index, cols_span, rows_span
                    ):
                        current = coords == (column_index, row_index)
                        cell_text, props = (
                            extract_props(SPAN_RE.sub("", value).strip(), "@")
                            if current
                            else ["", {}]
                        )
                        matrix[coords] = Cell(
                            cell_text,
                            skip=not current,
                            col_span=cols_span,
                            row_span=rows_span,
                            props=props,
                        )
            else:
                cell_text, props = extract_props(value, "@")
                matrix[column_index, row_index] = Cell(cell_text, props=props)
    return matrix


class SpanTable(Table):
    """
    Class holding information about a table represented in Markdown, supporting
    colspan and rowspan.
    """

    def __init__(
        self, headers: Iterable[str], records: Iterable[Iterable[str]]
    ) -> None:
        super().__init__(headers, records)
        self._matrix = get_matrix(self)

    @property
    def matrix(self) -> Matrix:
        return self._matrix



import logging
import xml.etree.ElementTree as etree
from dataclasses import dataclass

from neoteroi.markdown.align import Alignment
from neoteroi.markdown.images import build_icon_html

from .domain import Timeline, TimelineItem

logger = logging.getLogger("MARKDOWN")


@dataclass
class TimelineViewOptions:
    id: str = ""
    class_name: str = ""
    alternate: bool = False
    align: Alignment = Alignment.LEFT
    headings: bool = False


class TimelineHTMLBuilder:
    def __init__(self, options: TimelineViewOptions) -> None:
        self.options = options

    def get_class(self) -> str:
        return (
            "nt-timeline vertical "
            + self.options.align.value
            + (" alternate" if self.options.alternate else "")
            + (f" {self.options.class_name}" if self.options.class_name else "")
        )

    def get_dot_class(self, item: TimelineItem) -> str:
        return "nt-timeline-dot " + (item.key or "") + (" bigger" if item.icon else "")

    def build_html(self, parent, timeline: Timeline):
        root_props = {"class": self.get_class()}
        if self.options.id:
            root_props["id"] = self.options.id

        root_element = etree.SubElement(parent, "div", root_props)
        etree.SubElement(root_element, "div", {"class": "nt-timeline-before"})

        items_element = etree.SubElement(
            root_element, "div", {"class": "nt-timeline-items"}
        )

        for item in timeline.items:
            self.build_item_html(items_element, item)

        etree.SubElement(root_element, "div", {"class": "nt-timeline-after"})

    def build_icon_html(self, parent, item: TimelineItem):
        build_icon_html(parent, item.icon)

    def build_item_html(self, parent, item: TimelineItem):
        item_element = etree.SubElement(
            parent,
            "div",
            {"class": "nt-timeline-item" + (f" {item.key}" if item.key else "")},
        )

        title_element = etree.SubElement(
            item_element,
            "h3" if self.options.headings else "p",
            {"class": "nt-timeline-title"},
        )
        title_element.text = item.title

        if item.sub_title:
            sub_title_element = etree.SubElement(
                item_element, "span", {"class": "nt-timeline-sub-title"}
            )
            sub_title_element.text = item.sub_title

        if item.content:
            content_element = etree.SubElement(
                item_element, "p", {"class": "nt-timeline-content"}
            )
            content_element.text = item.content

        dot_element = etree.SubElement(
            item_element, "div", {"class": self.get_dot_class(item)}
        )

        self.build_icon_html(dot_element, item)



"""
This module provides an extension for a component to display chronological information
with style.

neoteroi.timeline

MIT License
Copyright (c) 2022 to present, Roberto Prevato
"""
from markdown import Extension

from neoteroi.markdown.align import aligment_from_props
from neoteroi.markdown.processors import EmbeddedBlockProcessor, SourceBlockProcessor
from neoteroi.markdown.utils import create_instance, create_instances

from .domain import Timeline, TimelineItem
from .html import TimelineHTMLBuilder, TimelineViewOptions


class BaseTimelineProcessor:
    @property
    def name(self) -> str:
        return "timeline"

    def _norm_props(self, props) -> None:
        if "class" in props:
            props["class_name"] = props["class"]
        props["align"] = aligment_from_props(props)

    def build_html(self, parent, obj, props) -> None:
        """Builds the HTML for the given input object."""
        if not isinstance(obj, list):
            raise TypeError("Expected a list of items describing timeline parts.")

        self._norm_props(props)
        builder = TimelineHTMLBuilder(create_instance(TimelineViewOptions, props))
        builder.build_html(parent, Timeline(create_instances(TimelineItem, obj)))


class TimelineEmbeddedProcessor(BaseTimelineProcessor, EmbeddedBlockProcessor):
    """
    Block processor that can render a timeline using data embedded in the Markdown.
    """


class TimelineSourceProcessor(BaseTimelineProcessor, SourceBlockProcessor):
    """
    Block processor that can render a timeline using data from a source outside of the
    Markdown (e.g. file, URL, database).
    """


class TimelineExtension(Extension):
    """Extension that includes timelines."""

    config = {
        "priority": [12, "The priority to be configured for the extension."],
    }

    def extendMarkdown(self, md):
        md.registerExtension(self)
        priority = self.getConfig("priority")

        md.parser.blockprocessors.register(
            TimelineEmbeddedProcessor(md.parser), "timeline", priority + 0.1
        )

        md.parser.blockprocessors.register(
            TimelineSourceProcessor(md.parser), "timeline-from-source", priority
        )



from pathlib import Path
from typing import Any

from .source import DataReader


def read_file(file_path: Path, encoding: str = "utf-8") -> str:
    with open(file_path, "rt", encoding=encoding) as source_file:
        return source_file.read()


class FileReader(DataReader):
    encoding = "utf-8"

    def test(self, source: str) -> bool:
        source_path = Path(source)
        return source_path.exists() and source_path.is_file()

    def read(self, source: str) -> Any:
        assert self.test(source)
        return read_file(Path(source), encoding=self.encoding)



import xml.etree.ElementTree as etree
from dataclasses import dataclass
from typing import Optional

from .utils import create_instance


@dataclass
class Image:
    url: str
    height: Optional[int] = None
    width: Optional[int] = None
    alt: Optional[str] = None

    @classmethod
    def from_obj(cls, obj):
        if isinstance(obj, str):
            return cls(obj)
        elif isinstance(obj, dict):
            return create_instance(cls, obj)

        raise TypeError("The given object is not of a supported type.")

    def get_props(self):
        props = {"src": self.url}

        if self.height is not None:
            props["height"] = str(self.height)

        if self.width is not None:
            props["width"] = str(self.width)

        if self.alt:
            props["alt"] = str(self.alt)

        return props


def build_image_html(parent, image: Image):
    etree.SubElement(parent, "img", image.get_props())


def build_icon_html(parent, icon):
    if not icon:
        return

    if "/" in icon:
        etree.SubElement(
            parent, "img", {"class": "icon", "src": icon, "alt": "step icon"}
        )
    elif "fa-" in icon:
        # Fontawesome
        etree.SubElement(parent, "i", {"class": f"{icon} icon"})
    else:
        # other icon - this integrates with other processors, like the one from
        # Material for MkDocs!
        span = etree.SubElement(parent, "span", {"class": "icon"})
        span.text = icon



"""
This module defines common processor classes used in the package. It defines classes to
handle common notations like:

[example(pros?)(content-source)]

and

::example:: (props?) ...
content
::/example::
"""

import logging
import re
import textwrap
import xml.etree.ElementTree as etree
from abc import ABC, abstractmethod
from typing import Iterable, List, Optional

from markdown.blockprocessors import BlockProcessor

from neoteroi.markdown import parse_props
from neoteroi.markdown.data.files import FileReader
from neoteroi.markdown.data.source import DataReader
from neoteroi.markdown.data.text import CSVParser, JSONParser, TextParser, YAMLParser
from neoteroi.markdown.data.web import HTTPDataReader

logger = logging.getLogger("MARKDOWN")


def find_closing_fragment_index(pattern: re.Pattern, blocks: List[str]) -> int:
    for index, block in enumerate(blocks):
        if pattern.search(block):
            return index
    return -1


def pop_to_index(items, index):
    """
    Pops elements from a source list, and yields them up to the given index,
    included.

    Example:
    pop_to_index([1, 2, 3, 4], 2) -> yields 1, 2, 3,
    removing them from the source list.
    """
    i = 0

    while i <= index:
        block = items.pop(0)
        yield block + "\n"
        i += 1


class BaseProcessor(ABC):
    parsers: Iterable[TextParser] = (YAMLParser(), JSONParser(), CSVParser())

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns the name of the tag that will be handled by this processor."""

    @abstractmethod
    def build_html(self, parent, obj, props) -> None:
        """Builds the HTML for the given input object."""

    def get_parsers(self, props):
        """
        Tries to get the best parser by tag property.
        """
        if props.get("yaml") is True:
            return [YAMLParser()]

        if props.get("json") is True:
            return [JSONParser()]

        if props.get("csv") is True:
            return [CSVParser()]

        return self.parsers

    def render_courtesy_error(self, parent, message: str):
        div = etree.SubElement(parent, "div", {"class": "nt-error"})
        p = etree.SubElement(div, "p")
        p.text = message

    def parse(self, text, props):
        for parser in self.get_parsers(props):
            try:
                return parser.parse(text)
            except (TypeError, ValueError) as parser_exc:
                logger.debug(
                    "The parser %s failed to parse the text.",
                    type(parser).__qualname__,
                    exc_info=parser_exc,
                )

        raise ValueError("The input text could not be parsed.")

    def render(self, parent, data, props):
        if isinstance(data, str):
            try:
                obj = self.parse(data, props)
            except ValueError:
                logger.exception("Could not parse the value of a %s block.", self.name)
                self.render_courtesy_error(
                    parent,
                    f"Could not parse the value of this {self.name} block. "
                    "Please correct the input.",
                )
                return
        else:
            obj = data

        try:
            self.build_html(parent, obj, props)
        except (ValueError, TypeError):
            logger.exception("Could not render a %s block.", self.name)
            self.render_courtesy_error(
                parent,
                f"Could not render a {self.name} block. Please correct the input.",
            )

    def get_match(self, pattern, blocks) -> Optional[re.Match]:
        first_block = blocks.pop(0)
        new_lines = []
        match = None

        for line in first_block.splitlines():
            match = pattern.match(line)
            if match:
                break
            else:
                new_lines.append(line)
        if new_lines:
            blocks.insert(0, "\n".join(new_lines))

        return match


class SourceBlockProcessor(BlockProcessor, BaseProcessor):
    """
    Base class for inline processors that read information from a file or URL source
    and handle it to create an HTML element, like:

    [timeline(./example.yaml)]

    [timeline(https://.../example.json)]
    """

    _pattern: Optional[re.Pattern] = None

    @property
    def pattern(self) -> re.Pattern:
        if self._pattern is None:
            self._pattern = re.compile(
                rf"(?P<indent>\s*)\[{self.name}\s?([^\(]*)\((.*?)\)\]", re.DOTALL
            )
        return self._pattern

    def test(self, parent, block) -> bool:
        if self.pattern.search(block):
            return True
        else:
            return False

    data_readers: Iterable[DataReader] = (FileReader(), HTTPDataReader())

    def get_data_reader(self, source: str) -> DataReader:
        if not source:
            raise ValueError("Missing source parameter.")

        for reader in self.data_readers:
            if reader.test(source):
                return reader

        logger.warning("[%s] could not resolve the source: %s", self.name, source)
        raise ValueError(
            f'[{self.name}] invalid source. The source "{source}" could not be '
            "resolved. If the source is a file, please verify the path."
        )

    def read_from_source(self, source: str):
        reader = self.get_data_reader(source)
        return reader.read(source)

    def run(self, parent, blocks):
        match = self.get_match(self.pattern, blocks)
        assert match is not None

        groups = match.groups()
        raw_props = groups[1]
        source = groups[2]

        if raw_props:
            props = parse_props(raw_props, bool_attrs=True)
        else:
            props = {}

        try:
            data = self.read_from_source(source)
        except ValueError as value_error:
            self.render_courtesy_error(parent, str(value_error))
        else:
            self.render(parent, data, props)


class EmbeddedBlockProcessor(BlockProcessor, BaseProcessor):
    """
    Base class for standard block processors that handle a fragment like:

    ::example:: format prop_1="value_1" prop_2="value_2" ...
    [CSV | JSON | YAML]
    ::/example::

    This class provides base functions to handle the Markdown integration and parsing of
    the input, concrete classes define how to build the HTML once the input has been
    parsed.

    The source of data in this case is always coming as a string.
    """

    _start_pattern: Optional[re.Pattern] = None
    _end_pattern: Optional[re.Pattern] = None

    @property
    def start_pattern(self) -> re.Pattern:
        if self._start_pattern is None:
            self._start_pattern = re.compile(
                rf"""(?P<indent>\s*)::{self.name}::([^\n]*)?""", re.DOTALL
            )
        return self._start_pattern

    @property
    def end_pattern(self) -> re.Pattern:
        if self._end_pattern is None:
            self._end_pattern = re.compile(rf"\s*::/{self.name}::\s*\n?")
        return self._end_pattern

    def test(self, parent, block) -> bool:
        if self.start_pattern.search(block):
            return True
        else:
            return False

    def find_closing_fragment_index(self, blocks) -> int:
        return find_closing_fragment_index(self.end_pattern, blocks)

    def get_content(self, relevant_blocks):
        raw_text = textwrap.dedent("".join(relevant_blocks))
        return self.end_pattern.sub("", self.start_pattern.sub("", raw_text))

    def run(self, parent, blocks):
        closing_block_index = self.find_closing_fragment_index(blocks)

        if closing_block_index == -1:
            # unclosed tag, ignore
            logger.warning(
                f"Unclosed ::{self.name}:: block - expected a ::/{self.name}:: block."
            )
            return False
        match = self.start_pattern.match(blocks[0])

        assert match is not None, "A match here is expected"

        props = parse_props(
            match.group(0).lstrip().replace(f"::{self.name}::", ""), bool_attrs=True
        )

        if closing_block_index == 0:
            # there is no carriage return, which is not ideal for
            # what we want to achieve
            single_block = blocks[0].lstrip()
            raw_text = textwrap.dedent(
                self.end_pattern.sub("", self.start_pattern.sub("", single_block))
            ).lstrip()
            blocks.pop(0)
        else:
            blocks.pop(0)
            raw_text = self.get_content(pop_to_index(blocks, closing_block_index - 1))

        self.render(parent, raw_text, props)



"""
Objects to describe alignment of elements.
"""
import logging
from enum import Enum

logger = logging.getLogger("MARKDOWN")


class Alignment(Enum):
    CENTER = "center"
    LEFT = "left"
    RIGHT = "right"


def aligment_from_props(props):
    if props.get("center") is True:
        return Alignment.CENTER

    if props.get("left") is True:
        return Alignment.LEFT

    if props.get("right") is True:
        return Alignment.RIGHT

    return try_parse_align(props.get("align", "left"))


def try_parse_align(align: str):
    try:
        return Alignment(align)
    except ValueError:
        logger.exception(
            "Invalid timeline align property: %s; ignoring this attribute", align
        )
        return Alignment.LEFT



"""
This module provides an implementation of cards.

neoteroi.cards

MIT License
Copyright (c) 2022 to present, Roberto Prevato
"""
from markdown import Extension

from neoteroi.markdown.images import Image
from neoteroi.markdown.processors import EmbeddedBlockProcessor, SourceBlockProcessor
from neoteroi.markdown.utils import create_instance, create_instances

from .domain import CardItem, Cards
from .html import CardsHTMLBuilder, CardsViewOptions


class BaseCardsProcessor:
    @property
    def name(self) -> str:
        return "cards"

    def norm_obj(self, obj):
        for item in obj:
            if "image" in item:
                item["image"] = Image.from_obj(item["image"])

    def build_html(self, parent, obj, props) -> None:
        """Builds the HTML for the given input object."""
        if not isinstance(obj, list):
            raise TypeError("Expected a list of items describing cards.")

        self.norm_obj(obj)
        builder = CardsHTMLBuilder(create_instance(CardsViewOptions, props))
        builder.build_html(parent, Cards(create_instances(CardItem, obj)))


class CardsEmbeddedProcessor(BaseCardsProcessor, EmbeddedBlockProcessor):
    """
    Block processor that can render cards using data embedded in the Markdown.
    """


class CardsSourceProcessor(BaseCardsProcessor, SourceBlockProcessor):
    """
    Block processor that can render cards using data from a source outside of the
    Markdown (e.g. file, URL, database).
    """


class CardsExtension(Extension):
    """Extension that includes cards."""

    config = {
        "priority": [12, "The priority to be configured for the extension."],
    }

    def extendMarkdown(self, md):
        md.registerExtension(self)
        priority = self.getConfig("priority")

        md.parser.blockprocessors.register(
            CardsEmbeddedProcessor(md.parser), "cards", priority + 0.1
        )

        md.parser.blockprocessors.register(
            CardsSourceProcessor(md.parser), "cards-from-source", priority
        )



"""
This module provides an extension to support Markdown tables with colspan and rowspan.

neoteroi.spantable

MIT License
Copyright (c) 2022 to present, Roberto Prevato
"""
import logging
import re
import xml.etree.ElementTree as etree
from typing import Dict, Optional

from markdown import Extension
from markdown.blockprocessors import BlockProcessor

from neoteroi.markdown import parse_props
from neoteroi.markdown.tables import read_table
from neoteroi.markdown.tables.spantable import Cell, SpanTable

logger = logging.getLogger("MARKDOWN")


class SpanTableProcessor(BlockProcessor):

    START_RE = re.compile(r"""(?P<indent>\s*)::spantable::[\w\s]*""", re.DOTALL)

    END_RE = re.compile(r"\s*:{2}end-spantable:{2}\s*\n?")

    def test(self, parent, block):
        if self.START_RE.search(block):
            return True
        else:
            return False

    def find_closing_fragment_index(self, blocks) -> int:
        for index, block in enumerate(blocks):
            if self.END_RE.search(block):
                return index
        return -1

    def pop_to_index(self, items, index):
        """
        Pops elements from a source list, and yields them up to the given index,
        included.

        Example:
        pop_to_index([1, 2, 3, 4], 2) -> yields 1, 2, 3,
        removing them from the source list.
        """
        i = 0

        while i <= index:
            block = items.pop(0)
            yield block + "\n"
            i += 1

    def read_first_table(self, blocks) -> Optional[SpanTable]:
        """
        Reads the first table in the given blocks, returning it.
        If a block does not contain a table, it is ignored.
        """
        for block in blocks:
            table = read_table(block, SpanTable)

            if table is not None:
                return table

    def run(self, parent, blocks):
        """
        Converts Markdown code to an HTML table, injecting it to the given parent.
        """
        closing_block_index = self.find_closing_fragment_index(blocks)

        if closing_block_index == -1:
            # unclosed tag, ignore
            logger.warning(
                "Unclosed ::spantable:: block - expected a ::end-spantable::."
            )
            return False

        props = parse_props(blocks[0])

        table_blocks = list(self.pop_to_index(blocks, closing_block_index))
        span_table = self.read_first_table(table_blocks)
        assert span_table is not None, "A table is expected if test() -> True"

        div = etree.SubElement(parent, "div", {"class": "span-table-wrapper"})
        table = etree.SubElement(div, "table", self._get_table_props(props))

        self._handle_caption(table, props)
        self._set_table(span_table, table)

    def _handle_caption(self, table_element: etree.Element, props: Dict[str, str]):
        caption = props.get("caption")
        if caption:
            caption_el = etree.SubElement(table_element, "caption", {})
            caption_el.text = caption

    def _get_table_props(self, props):
        data = {"class": "span-table"}

        html_class = props.get("class")
        if html_class:
            data["class"] = f"span-table {html_class}"

        return data

    def _set_table(self, span_table: SpanTable, table_element: etree.Element):
        for row in span_table.matrix.rows:
            tr = etree.SubElement(table_element, "tr")

            for cell in row:
                assert isinstance(cell, Cell)
                if cell.skip:
                    continue
                props = {}

                if cell.html_class:
                    props["class"] = cell.html_class

                if cell.col_span > 1:
                    props["colspan"] = str(cell.col_span)

                if cell.row_span > 1:
                    props["rowspan"] = str(cell.row_span)

                td = etree.SubElement(tr, "td", props)
                td.text = cell.text


class SpanTableExtension(Extension):
    """Extension that includes SpanTables (tables supporting rowspan and colspan)."""

    def extendMarkdown(self, md):
        md.registerExtension(self)

        md.parser.blockprocessors.register(
            SpanTableProcessor(md.parser), "spantable", 106
        )


def make_extension(*args, **kwargs):
    return SpanTableExtension(*args, **kwargs)



from dataclasses import dataclass
from typing import List, Optional

from neoteroi.markdown.images import Image


@dataclass
class CardItem:
    title: str
    url: Optional[str] = None
    content: Optional[str] = None
    icon: Optional[str] = None
    key: Optional[str] = None
    image: Optional[Image] = None


@dataclass
class Cards:
    items: List[CardItem]



"""
This module defines a base class for types that can read text from a source.
"""
from abc import ABC, abstractmethod
from typing import Any


class DataReader(ABC):
    @abstractmethod
    def test(self, source: str) -> bool:
        """
        Returns a value indicating whether this DataReader can read from a source
        described by the given value.
        """

    @abstractmethod
    def read(self, source: str) -> Any:
        """Returns data read from a source."""



import re
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, TypeVar

_TABLE_LINE_PATTERN = re.compile(r"\s?\|([^\|]+)")
_TABLE_SEPARATOR_LINE_PATTERN = re.compile(r"^[-:\s\|]*$")


class Matrix:
    """
    A matrix of rows and columns, with mutable values.
    """

    def __init__(
        self, columns_count: int, rows_count: int, initial_value: Any = None
    ) -> None:
        self._columns_count = columns_count
        self._rows_count = rows_count
        self._rows = tuple(
            [[initial_value for _ in range(columns_count)] for _ in range(rows_count)]
        )

    @property
    def rows(self) -> Tuple[List[Any], ...]:
        return self._rows

    def __str__(self) -> str:
        return f"<Matrix {str(self.rows)}>"

    def __repr__(self) -> str:
        return f"Matrix({self._columns_count}, {self._rows_count})"

    def __iter__(self):
        yield from self.rows

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise ValueError("Slice length out of boundary.")
            col, row = key
            self.rows[row][col] = value
            return

        raise ValueError("Expected a tuple of values (x, y).")

    def __getitem__(self, key):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise ValueError("Slice length out of boundary.")
            col, row = key
            return self.rows[row][col]

        return self.rows[key]


class Table:
    """
    Represents a read-only table with information obtained from Markdown.
    """

    def __init__(
        self, headers: Iterable[str], records: Iterable[Iterable[str]]
    ) -> None:
        self._headers = tuple(headers)
        self._records = tuple(tuple(record) for record in records)
        self._headers_indexes = {header: i for i, header in enumerate(self._headers)}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.headers}, "
            f"len: {len(self)} id: {id(self)})"
        )

    @property
    def headers(self) -> Tuple[str, ...]:
        return self._headers

    @property
    def records(self) -> Tuple[Tuple[str, ...], ...]:
        return self._records

    def items(self) -> Iterable[Dict[str, str]]:
        properties = {index: header for index, header in enumerate(self.headers)}

        for record in self.records:
            yield {properties[index]: value for index, value in enumerate(record)}

    def __iter__(self):
        yield from self.records

    def __len__(self):
        return len(self.records)

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return self.records[key]

        if isinstance(key, tuple):
            if len(key) > 2:
                raise ValueError("Slice length out of boundary.")
            col, row = key
            return self.records[row][col]

        if key not in self.headers:
            raise KeyError(key)
        index = self._headers_indexes[key]
        # TODO: handle missing values if a row has less values (this should not happen)
        return tuple(record[index] for record in self.records)


T = TypeVar("T", bound=Table)


def read_table(markdown: str, cls: Type[T] = Table) -> Optional[T]:
    headers: List[str] = []
    records: List[List[str]] = []

    for line in markdown.splitlines():
        if _TABLE_SEPARATOR_LINE_PATTERN.match(line):
            continue

        matches = list(_TABLE_LINE_PATTERN.finditer(line))

        if matches:
            row = [match.group(1).strip() for match in matches]
            if not headers:
                headers = row
            else:
                records.append(row)

    return cls(headers, records) if headers else None



import calendar
import re
from dataclasses import dataclass
from datetime import date, datetime
from typing import Dict, Iterable, List, Tuple

from dateutil import rrule
from dateutil.relativedelta import relativedelta

_LASTS_PATTERN = re.compile(r"(?P<amount>\d+)\s?(?P<unit>\w+)")


MONTHS: List[Tuple[int, str]] = [
    (1, "January"),
    (2, "February"),
    (3, "March"),
    (4, "April"),
    (5, "May"),
    (6, "June"),
    (7, "July"),
    (8, "August"),
    (9, "September"),
    (10, "October"),
    (11, "November"),
    (12, "December"),
]

QUARTERS_BY_MONTH_NUMBER: Dict[int, int] = {
    1: 1,
    2: 1,
    3: 1,
    4: 2,
    5: 2,
    6: 2,
    7: 3,
    8: 3,
    9: 3,
    10: 4,
    11: 4,
    12: 4,
}


_MIN_MONTHS_BY_QUARTER_NUMBER: Dict[int, int] = {
    1: 1,
    2: 4,
    3: 7,
    4: 10,
}


_MAX_MONTHS_BY_QUARTER_NUMBER: Dict[int, int] = {
    1: 3,
    2: 6,
    3: 9,
    4: 12,
}


@dataclass(frozen=True)
class Quarter:
    year: int
    number: int

    @property
    def min_month(self) -> int:
        return _MIN_MONTHS_BY_QUARTER_NUMBER[self.number]

    @property
    def max_month(self) -> int:
        return _MAX_MONTHS_BY_QUARTER_NUMBER[self.number]

    @property
    def min_month_date(self) -> date:
        return date(self.year, self.min_month, 1)

    @property
    def max_month_date(self) -> date:
        return date(self.year, self.max_month, 1)

    def iter_dates(self) -> Iterable[date]:
        for value in iter_months_between_dates(
            self.min_month_date, self.max_month_date
        ):
            yield value.date()


class InvalidLastsValue(ValueError):
    def __init__(self, value: str) -> None:
        super().__init__(f"Invalid lasts value: {value}.")


class UnsupportedLastsUnit(ValueError):
    def __init__(self, value: str, unit: str) -> None:
        super().__init__(f"Unsupported lasts unit: {unit} in {value}.")


def parse_lasts(value: str) -> relativedelta:
    """
    Parses a lasts value into an instance of timedelta.

    For example:
    "20 days" -> timedelta(days=20)
    """
    if not value:
        raise ValueError("Missing input value")

    match = _LASTS_PATTERN.match(value)
    if not match:
        raise InvalidLastsValue(value)
    groups = match.groupdict()

    try:
        amount = int(groups["amount"])
    except ValueError:
        raise InvalidLastsValue(value)

    unit = groups["unit"].lower()

    if unit in {"year", "years", "y"}:
        return relativedelta(days=amount * 365)

    if unit in {"month", "months", "m"}:
        return relativedelta(months=amount)

    if unit in {"week", "weeks", "w"}:
        return relativedelta(weeks=amount)

    if unit in {"day", "days", "d"}:
        return relativedelta(days=amount)

    if unit in {"hour", "hours", "h"}:
        return relativedelta(hours=amount)

    raise UnsupportedLastsUnit(value, unit)


def iter_years_between_dates(min_date: date, max_date: date) -> Iterable[int]:
    if max_date < min_date:
        raise ValueError("max_date must be smaller than min_date.")

    year = min_date.year
    max_year = max_date.year

    yield year

    while year < max_year:
        year += 1
        yield year


def iter_months_between_dates_of_year(
    min_date: date, max_date: date, year: int
) -> Iterable[date]:
    if min_date.year > year:
        raise ValueError("Year out of boundary")
    start_date = min_date.replace(day=1)
    end_date = get_last_day_of_month(max_date)

    for month_date in rrule.rrule(rrule.MONTHLY, dtstart=start_date, until=end_date):
        if month_date.year == year:
            yield month_date.date()
        if month_date.year > year:
            break


def iter_months_between_dates(min_date: date, max_date: date) -> Iterable[datetime]:
    """
    Iterates month dates between two dates.
    """
    start_date = min_date.replace(day=1)
    end_date = max_date.replace(day=1)

    for month_date in rrule.rrule(rrule.MONTHLY, dtstart=start_date, until=end_date):
        yield month_date


def iter_days_between_dates(min_date: date, max_date: date) -> Iterable[date]:
    """
    Iterates dates between two dates.
    """
    start_date = min_date.replace(day=1)
    end_date = get_last_day_of_month(max_date)

    for day_date in rrule.rrule(rrule.DAILY, dtstart=start_date, until=end_date):
        yield day_date


def iter_quarters_between_dates(min_date: date, max_date: date) -> Iterable[Quarter]:
    """
    Iterates quarters numbers between two dates.
    """
    current_value: int = -1
    for month_date in iter_months_between_dates(min_date, max_date):
        value = QUARTERS_BY_MONTH_NUMBER[month_date.month]

        if value != current_value:
            current_value = value
            yield Quarter(month_date.year, value)


def iter_weeks_between_dates(
    min_date: date, max_date: date
) -> Iterable[Tuple[int, date]]:
    current_week_year, current_week_number, _ = min_date.isocalendar()
    current_week_date = date.fromisocalendar(current_week_year, current_week_number, 1)
    end_date = get_last_day_of_month(max_date)
    week_date: datetime

    for week_date in rrule.rrule(
        rrule.WEEKLY, dtstart=current_week_date, until=end_date
    ):
        yield week_date.isocalendar()[1], week_date.date()


def get_next_week_date(week_number, week_date) -> date:
    try:
        return date.fromisocalendar(week_date.year, week_number + 1, 1)
    except ValueError:
        return date.fromisocalendar(week_date.year + 1, 1, 1)


def get_first_day_of_month(value: date) -> date:
    """
    Returns the date of the first day of a month from a given date.
    """
    return date(value.year, value.month, 1)


def get_last_day_of_month(value: date) -> date:
    """
    Returns the date of the last day of a month from a given date.
    """
    return date(
        value.year, value.month, calendar.monthrange(value.year, value.month)[1]
    )


def date_delta(value1: date, value2: date):
    """
    Returns the delta between the date components of two dates, supporting
    datetimes.
    Note: a simple substraction causes TypeError if any of the two dates
    is a datetime.
    """
    if isinstance(value1, datetime):
        value1 = value1.date()
    if isinstance(value2, datetime):
        value2 = value2.date()
    return value1 - value2



"""
This module defines a base class for types that can deserialize text into Python
objects, and implementations for common formats.
"""
import csv
import json
from abc import ABC, abstractmethod
from io import StringIO
from typing import Any

import yaml


class TextParser(ABC):
    """Base class for classes that can deserialize text into Python objects."""

    @abstractmethod
    def parse(self, text) -> Any:
        """Parses a text into a more complex object."""


class YAMLParser(TextParser):
    def parse(self, text) -> Any:
        return yaml.safe_load(text)


class JSONParser(TextParser):
    def parse(self, text) -> Any:
        return json.loads(text)


class CSVParser(TextParser):
    def get_reader(self, csv_file):
        return csv.DictReader(csv_file, delimiter=",")

    def parse(self, text) -> Any:
        with StringIO(text) as string_io:
            reader = self.get_reader(string_io)
            return [record for record in reader]



"""
This module provide a plugin that can be used to generate Markdown
from OpenAPI Documentation files, using essentials-openapi

    https://github.com/Neoteroi/essentials-openapi

The markdown requires by default

neoteroi.mkdocsoad
"""
import re

from mkdocs.config.config_options import Type
from mkdocs.plugins import BasePlugin
from openapidocs.mk.v3 import OpenAPIV3DocumentationHandler
from openapidocs.utils.source import read_from_source


class MkDocsOpenAPIDocumentationPlugin(BasePlugin):
    config_scheme = (("use_pymdownx", Type(bool, default=False)),)

    rx = re.compile(r"\[OAD\(([^\)]+)\)\]")

    def _get_style(self) -> str:
        return "MKDOCS" if self.config.get("use_pymdownx", False) else "MARKDOWN"

    def _replacer(self, match) -> str:
        source = match.group(1).strip("'\"")
        data = read_from_source(source)

        handler = OpenAPIV3DocumentationHandler(
            data, style=self._get_style(), source=source
        )
        return handler.write()

    def on_page_markdown(self, markdown, *args, **kwargs):
        """
        Replaces the tag [OAD(...)] in markdown with markdown generated from an
        OpenAPI Documentation, using essentials-openapi

        https://github.com/Neoteroi/essentials-openapi
        """
        if "[OAD(" in markdown:
            return self.rx.sub(self._replacer, markdown)
        return markdown



import os
from typing import Any

import httpx

from .source import DataReader

HTTPX_DISABLE_SSL_VERIFY = bool(os.environ.get("HTTPX_DISABLE_SSL_VERIFY"))

default_http_client = httpx.Client(verify=not HTTPX_DISABLE_SSL_VERIFY, timeout=20)


class FailedRequestError(Exception):
    def __init__(self, message) -> None:
        super().__init__(
            f"Failed request: {message}. "
            "Inspect the inner exception (__context__) for more information."
        )

    @property
    def inner_exception(self):
        return self.__context__


def ensure_success(response: httpx.Response) -> None:
    if response.status_code < 200 or response.status_code > 399:
        raise FailedRequestError(
            "Response status does not indicate success: "
            f"{response.status_code} {response.reason_phrase}"
        )


class HTTPDataReader(DataReader):
    http_client: httpx.Client = default_http_client

    def test(self, source: str) -> bool:
        source_lower = source.lower()
        return source_lower.startswith("http://") or source_lower.startswith("https://")

    def get(self, url: str) -> httpx.Response:
        try:
            return self.http_client.get(url)
        except httpx.HTTPError as http_error:  # pragma: no cover
            raise FailedRequestError(str(http_error)) from http_error

    def read(self, source: str) -> Any:
        assert self.test(source)
        response = self.get(source)

        ensure_success(response)
        return response.text



"""
Common utilities for the whole package.
"""
from dataclasses import fields, is_dataclass

_FIELDS = {}


def _get_fields(cls):
    try:
        return _FIELDS[cls]
    except KeyError:
        _FIELDS[cls] = {f.name for f in fields(cls)}
        return _FIELDS[cls]


def _norm_key(key: str) -> str:
    return key.replace("-", "_").lower()


def create_instance(cls, props, field_names=None):
    """
    Creates an instance of a given dataclass type, ignoring extra properties.
    """
    if not is_dataclass(cls):
        raise ValueError("The given type is not a dataclass")
    if field_names is None:
        field_names = _get_fields(cls)
    return cls(
        **{_norm_key(k): v for k, v in props.items() if _norm_key(k) in field_names}
    )


def create_instances(cls, items):
    """
    Creates a list of instances of a given dataclass type, ignoring extra properties
    on single items.
    """
    field_names = _get_fields(cls)
    return [create_instance(cls, item, field_names) for item in items]



"""
This package contains common utilities for markdown used across the various plugins
for Markdown.
"""
import re
from typing import Dict, Tuple

_PROPS_RE = re.compile(
    r"""\s?((?P<name>[^\s\=]+)=(?P<quot>"|')(?P<value>[^\"\']+)(?P=quot))""",
    re.DOTALL | re.MULTILINE,
)


def parse_props(
    line: str, prefix: str = "", bool_attrs: bool = False
) -> Dict[str, str]:
    """
    Parses a line describing properties, in this form:

    prop1="value1" prop2="value2"

    It also parses boolean attributes, if bool_attrs is set to true:

    id="example" class="foo another" multiple

    into a dictionary of names and values. It accepts an optional prefix, to filter
    properties names.
    """
    props = {}

    for match in _PROPS_RE.findall(line):
        name: str = match[1]
        value: str = match[3]
        if prefix:
            if name.startswith(prefix):
                props[name[len(prefix) :]] = value
        else:
            props[name] = value

    if bool_attrs:
        for word in _PROPS_RE.sub("", line).split():
            if "=" in word:
                name, value = word.split("=")
                props[name] = value
            else:
                props[word] = True

    return props


def extract_props(line: str, prefix: str = "") -> Tuple[str, Dict[str, str]]:
    """
    Extracts properties like the `parse_props` function, but
    also returns the original line with properties removed.
    """
    props = parse_props(line, prefix)
    return _PROPS_RE.sub("", line).strip(), props



from dataclasses import dataclass
from datetime import date, datetime
from typing import Iterable, List, Optional, Union

from dateutil.parser import parse as parse_date

from neoteroi.projects.timeutil import parse_lasts


@dataclass(frozen=True)
class Event:
    title: str
    description: Optional[str] = None
    time: Optional[datetime] = None
    icon: Optional[str] = None

    def __post_init__(self):
        # Note: datetimes are currently not supported, only the date component is kept
        if isinstance(self.time, datetime):
            object.__setattr__(self, "time", self.time.date())

    @classmethod
    def from_obj(cls, obj):
        return cls(
            obj.get("title") or "",
            obj.get("description") or "",
            _parse_optional_datetime(obj.get("time")),
            icon=obj.get("icon"),
        )


def _parse_optional_datetime(value: Union[None, datetime, str]) -> Optional[datetime]:
    if isinstance(value, date):
        # YAML parses dates automatically
        return value

    if value:
        return parse_date(value)
    return None


def _parse_optional_date(value: Union[None, date, str]) -> Optional[date]:
    if isinstance(value, date):
        # YAML parses dates automatically
        return value

    if value:
        return parse_date(value)
    return None


@dataclass(frozen=True)
class Activity:
    title: str
    start: Optional[date] = None
    end: Optional[date] = None
    description: Optional[str] = None
    activities: Optional[List["Activity"]] = None
    events: Optional[List[Event]] = None

    def __post_init__(self):
        # Note: datetimes are currently not supported, only the date component is kept
        if isinstance(self.start, datetime):
            object.__setattr__(self, "start", self.start.date())
        if isinstance(self.end, datetime):
            object.__setattr__(self, "end", self.end.date())

    def iter_activities(self, include_self: bool = True) -> Iterable["Activity"]:
        """
        Yields self and all descendant activities.
        """
        if include_self:
            yield self

        if self.activities:
            for activity in self.activities:
                yield from activity.iter_activities()

    def iter_dates(self) -> Iterable[date]:
        if self.events:
            for event in self.events:
                if event.time:
                    yield event.time
        if self.activities:
            for activity in self.activities:
                if activity.start:
                    yield activity.start
                if activity.end:
                    yield activity.end

    def _get_event_date(self, fn) -> Optional[date]:
        if self.events:
            dates = [event.time for event in self.events if event.time is not None]
            return fn(dates) if dates else None
        return None

    def get_first_event_date(self) -> Optional[date]:
        return self._get_event_date(min)

    def get_last_event_date(self) -> Optional[date]:
        return self._get_event_date(max)

    def get_overall_start(self) -> Optional[date]:
        """
        Returns the start date of this activity, including all sub and descendants
        activities.
        """
        if self.activities:
            starts = [
                start
                for start in (
                    activity.get_overall_start() for activity in self.activities
                )
                if start is not None
            ]
            return min(starts) if starts else None
        first_event_date = self.get_first_event_date()
        if first_event_date and self.start:
            return min(first_event_date, self.start)
        return self.start or first_event_date

    def get_overall_end(self) -> Optional[date]:
        """
        Returns the end date of this activity, including all sub and descendants
        activities.
        """
        if self.activities:
            ends = [
                end
                for end in (activity.get_overall_end() for activity in self.activities)
                if end is not None
            ]
            return max(ends) if ends else None
        last_event_date = self.get_last_event_date()
        if last_event_date and self.end:
            return min(last_event_date, self.end)
        return self.end or last_event_date

    @classmethod
    def from_obj(cls, obj):
        if isinstance(obj, str):
            return cls(title=obj)
        if not isinstance(obj, dict):
            raise TypeError("Expected a str or a dict.")

        title = obj.get("title", "")
        description = obj.get("description")
        start = _parse_optional_date(obj.get("start"))
        end = _parse_optional_date(obj.get("end"))
        lasts = obj.get("lasts")
        child_activities = obj.get("activities")
        events = obj.get("events")
        if lasts:
            if start is None:
                start = date.today()
            end = start + parse_lasts(lasts)

        return cls(
            title,
            start,
            end,
            description=description,
            activities=[Activity.from_obj(item) for item in child_activities]
            if child_activities
            else None,
            events=[Event.from_obj(item) for item in events] if events else None,
        )


@dataclass(frozen=True)
class Plan(Activity):
    def iter_activities(self) -> Iterable["Activity"]:
        """
        Yields all descendant activities.
        """

        if self.activities:
            for activity in self.activities:
                yield from activity.iter_activities()

    @classmethod
    def from_obj(cls, obj):
        if isinstance(obj, list):
            return cls(
                title="Plan", activities=[Activity.from_obj(item) for item in obj]
            )

        if not isinstance(obj, dict):
            raise TypeError("Expected a list or a dict.")

        activities = obj.get("activities")
        events = obj.get("events")

        return cls(
            title=obj.get("title") or "Plan",
            activities=[Activity.from_obj(item) for item in activities]
            if activities
            else [],
            events=[Event.from_obj(item) for item in events] if events else [],
        )



from setuptools import setup


def readme():
    with open("README.md", mode="rt", encoding="utf8") as readme:
        return readme.read()


setup(
    name="neoteroi-mkdocs",
    version="0.0.6",
    description="Plugins for MkDocs and Python Markdown",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/Neoteroi/mkdocs-plugins",
    author="RobertoPrevato",
    author_email="roberto.prevato@gmail.com",
    keywords="MkDocs OpenAPI Swagger Markdown plugins extensions documentation",
    license="MIT",
    packages=[
        "neoteroi.markdown",
        "neoteroi.markdown.data",
        "neoteroi.markdown.tables",
        "neoteroi.mkdocsoad",
        "neoteroi.spantable",
        "neoteroi.timeline",
        "neoteroi.cards",
        "neoteroi.projects",
        "neoteroi.projects.gantt",
    ],
    install_requires=["essentials-openapi[full]", "mkdocs~=1.3.1", "httpx~=0.22.0"],
    entry_points={
        "mkdocs.plugins": [
            "neoteroi.mkdocsoad = neoteroi.mkdocsoad:MkDocsOpenAPIDocumentationPlugin",
        ],
        "markdown.extensions": [
            "neoteroi.spantable = neoteroi.spantable:SpanTableExtension",
            "neoteroi.timeline = neoteroi.timeline:TimelineExtension",
            "neoteroi.cards = neoteroi.cards:CardsExtension",
            "neoteroi.projects = neoteroi.projects:ProjectsExtension",
        ],
    },
    include_package_data=True,
    zip_safe=False,
)



import logging
import xml.etree.ElementTree as etree
from dataclasses import dataclass

from neoteroi.markdown.images import build_image_html

from .domain import CardItem, Cards

logger = logging.getLogger("MARKDOWN")


@dataclass
class CardsViewOptions:
    id: str = ""
    cols: int = 3
    image_bg: bool = False

    def __post_init__(self):
        if isinstance(self.cols, str):
            self.cols = int(self.cols)


class CardsHTMLBuilder:
    def __init__(self, options: CardsViewOptions) -> None:
        self.options = options

    @property
    def use_image_tags(self) -> bool:
        return not self.options.image_bg

    def get_item_props(self, item: CardItem):
        if item.key:
            item_props = {"class": f"nt-card {item.key}"}
        else:
            item_props = {"class": "nt-card"}

        return item_props

    def build_html(self, parent, cards: Cards):
        root_element = etree.SubElement(
            parent, "div", {"class": f"nt-cards nt-grid cols-{self.options.cols}"}
        )

        for item in cards.items:
            self.build_item_html(root_element, item)

    def build_item_html(self, parent, item: CardItem):
        item_element = etree.SubElement(parent, "div", self.get_item_props(item))

        if item.url:
            first_child = etree.SubElement(item_element, "a", {"href": item.url})
        else:
            first_child = etree.SubElement(
                item_element, "div", {"class": "nt-card-wrap"}
            )

        wrapper_element = etree.SubElement(first_child, "div", {})

        self.build_image_html(wrapper_element, item)

        text_wrapper = etree.SubElement(
            wrapper_element, "div", {"class": "nt-card-content"}
        )

        title_element = etree.SubElement(text_wrapper, "p", {"class": "nt-card-title"})
        title_element.text = item.title

        if item.content:
            content_element = etree.SubElement(
                text_wrapper, "p", {"class": "nt-card-text"}
            )
            content_element.text = item.content

    def build_image_html(self, wrapper_element, item: CardItem):
        if not item.image:
            return

        if self.use_image_tags:
            build_image_html(
                etree.SubElement(
                    wrapper_element, "div", {"class": "nt-card-image tags"}
                ),
                item.image,
            )
        else:
            etree.SubElement(
                wrapper_element,
                "div",
                {
                    "class": "nt-card-image",
                    "style": f"background-image: url('{item.image.url}')",
                },
            )



"""
This module provides a control to render Gantt diagrams.

neoteroi.gantt

MIT License
Copyright (c) 2022 to present, Roberto Prevato
"""
from neoteroi.markdown.processors import EmbeddedBlockProcessor, SourceBlockProcessor
from neoteroi.markdown.utils import create_instance

from ..domain import Plan
from .html import GanttHTMLBuilder, GanttViewOptions


class BaseGanttProcessor:
    @property
    def name(self) -> str:
        return "gantt"

    def build_html(self, parent, obj, props) -> None:
        """Builds the HTML for the given input object."""
        if not isinstance(obj, list):
            raise TypeError("Expected a list of items describing Gantt.")

        builder = GanttHTMLBuilder(
            Plan.from_obj(obj), create_instance(GanttViewOptions, props)
        )
        builder.build_html(parent)


class GanttEmbeddedProcessor(BaseGanttProcessor, EmbeddedBlockProcessor):
    """
    Block processor that can render a Gantt diagram using data embedded in the Markdown.
    """


class GanttSourceProcessor(BaseGanttProcessor, SourceBlockProcessor):
    """
    Block processor that can render a Gantt diagram using data from a source outside of
    the Markdown (e.g. file, URL, database).
    """


def register_extension(md, priority):
    md.parser.blockprocessors.register(
        GanttEmbeddedProcessor(md.parser), "gantt", priority + 0.1
    )

    md.parser.blockprocessors.register(
        GanttSourceProcessor(md.parser), "gantt-from-source", priority
    )



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class TimelineItem:
    title: str
    content: Optional[str] = None
    sub_title: Optional[str] = None
    icon: Optional[str] = None
    key: Optional[str] = None


@dataclass
class Timeline:
    items: List[TimelineItem]



import logging
import sys
import xml.etree.ElementTree as etree
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta
from enum import Enum
from typing import List, Union
from uuid import uuid4

from neoteroi.markdown.images import build_icon_html

from ..domain import Activity, Event, Plan
from ..timeutil import (
    date_delta,
    get_first_day_of_month,
    get_last_day_of_month,
    get_next_week_date,
    iter_days_between_dates,
    iter_months_between_dates,
    iter_months_between_dates_of_year,
    iter_quarters_between_dates,
    iter_weeks_between_dates,
    iter_years_between_dates,
)

logger = logging.getLogger("MARKDOWN")


class VerticalLinesPace(Enum):
    MONTHLY = "MONTHLY"
    WEEKLY = "WEEKLY"


@dataclass
class GanttViewOptions:
    id: str = ""
    month_width: float = 150
    whole_years: bool = False
    no_groups: bool = False
    no_years: bool = False
    no_weeks: bool = False
    no_quarters: bool = False
    no_days: bool = False
    pastello: bool = False
    vlines_pace: VerticalLinesPace = VerticalLinesPace.MONTHLY

    def __post_init__(self):
        if isinstance(self.month_width, str):
            self.month_width = float(self.month_width)
        if isinstance(self.vlines_pace, str):
            self.vlines_pace = VerticalLinesPace(self.vlines_pace.upper())
        elif self.month_width >= 500:
            self.vlines_pace = VerticalLinesPace.WEEKLY

    @property
    def day_width(self) -> float:
        return self.month_width / 30


@dataclass
class VerticalLine:
    left: float


@dataclass
class BuildContext:
    vertical_lines: List[VerticalLine] = field(default_factory=list)
    styles: List[str] = field(default_factory=list)


class GanttHTMLBuilder:
    def __init__(self, plan: Plan, options: GanttViewOptions) -> None:
        self._id = uuid4()
        self._plan = plan
        self._options = options

        self._min_date = plan.get_overall_start() or date.today()
        self._max_date = plan.get_overall_end() or date.today().replace(
            year=self._min_date.year + 1
        )

        if options.whole_years:
            self._scale_min_date = date(self._min_date.year, 1, 1)
            self._scale_max_date = date(self._max_date.year, 12, 31)
        else:
            self._scale_min_date = get_first_day_of_month(self._min_date)
            self._scale_max_date = get_last_day_of_month(self._max_date)

        self._months_count = len(
            list(iter_months_between_dates(self._scale_min_date, self._scale_max_date))
        )

    @property
    def plan(self) -> Plan:
        return self._plan

    @property
    def options(self) -> GanttViewOptions:
        return self._options

    @property
    def client_width(self) -> int:
        return int(self.options.month_width * self._months_count)

    @property
    def id(self) -> str:
        if self.options.id:
            return self.options.id

        return f"nt-plan-{str(self._id)}"

    def _get_root_class(self) -> str:
        classes = ["nt-plan-root"]
        if self.options.no_groups:
            classes.append("no-groups")
        if self.options.pastello:
            classes.append("nt-pastello")

        return " ".join(classes)

    def build_html(self, parent):
        context = BuildContext()
        root_element = etree.SubElement(
            parent, "div", {"class": self._get_root_class(), "id": self.id}
        )

        plan_element = etree.SubElement(root_element, "div", {"class": "nt-plan"})
        inner_element = etree.SubElement(
            plan_element, "div", {"class": "nt-plan-inner"}
        )

        self._build_periods_html(inner_element, context)
        self._build_groups_html(inner_element, context)
        self._build_style(root_element, context)

    def _build_style(self, parent, context: BuildContext):
        if not context.styles:
            return
        style_element = etree.SubElement(parent, "style")
        style_element.text = "\n".join(context.styles)

    def _build_periods_html(self, parent, context: BuildContext):
        periods_element = etree.SubElement(parent, "div", {"class": "nt-plan-periods"})
        min_date = self._min_date
        max_date = self._max_date

        if min_date and max_date:
            self._build_years(periods_element)

            if not self.options.no_quarters:
                self._build_quarters(periods_element)
            self._build_months(periods_element, context)
            self._build_weeks(periods_element, context)

            if self.options.day_width > 20 and not self.options.no_days:
                self._build_days(periods_element)

    def _build_years(self, periods_element):
        if self.options.no_years:
            return

        min_date = self._scale_min_date
        max_date = self._scale_max_date

        years_element = etree.SubElement(periods_element, "div", {"class": "years"})
        for i, year in enumerate(iter_years_between_dates(min_date, max_date)):
            # it can happen that a year has just a few months visible on the scale!!
            # therefore it is necessary to know also the month, to generate a proper
            # scale
            attributes = {"class": f"year year-{i} year-{year}"}
            if min_date.year != max_date.year:
                year_width = sum(
                    self._get_month_width(month_date)
                    for month_date in iter_months_between_dates_of_year(
                        min_date, max_date, year
                    )
                    if self._scale_max_date >= month_date >= self._scale_min_date
                )
                attributes["style"] = f"width: {year_width}px;"
            year_element = etree.SubElement(years_element, "div", attributes)
            year_element.text = str(year)

    def _build_quarters(self, periods_element):
        min_date = self._scale_min_date
        max_date = self._scale_max_date
        quarters_element = etree.SubElement(
            periods_element, "div", {"class": "quarters"}
        )
        for quarter in iter_quarters_between_dates(min_date, max_date):
            i = quarter.number
            attributes = {
                "class": f"quarter quarter-{i}",
                "title": f"{quarter.year} Q{quarter.number}",
            }

            quarter_width = sum(
                self._get_month_width(month_date)
                for month_date in quarter.iter_dates()
                if self._scale_max_date >= month_date >= self._scale_min_date
            )
            attributes["style"] = f"width: {quarter_width}px;"

            quarter_element = etree.SubElement(quarters_element, "div", attributes)
            quarter_element.text = f"Q{i}"

    def _build_months(self, periods_element, context: BuildContext):
        min_date = self._scale_min_date
        max_date = self._scale_max_date

        total_width: float = 0
        months_element = etree.SubElement(periods_element, "div", {"class": "months"})

        for month_date in iter_months_between_dates(min_date, max_date):
            name = month_date.strftime("%B").lower()
            month_width = self._get_month_width(month_date)
            month_element = etree.SubElement(
                months_element,
                "div",
                {
                    "class": f"month month-{name}",
                    "style": f"width: {month_width}px;",
                },
            )
            month_element.text = month_date.strftime("%B %Y")

            # add vertical line
            if self.options.vlines_pace == VerticalLinesPace.MONTHLY:
                total_width += month_width
                context.vertical_lines.append(VerticalLine(round(total_width, 2)))

    def _get_month_width(self, month_date) -> float:
        last_day_of_month = get_last_day_of_month(month_date).day

        if last_day_of_month == 30:
            return self.options.month_width

        return round(self.options.day_width * last_day_of_month, 2)

    def _build_weeks(self, periods_element, context: BuildContext):
        if self.options.no_weeks:
            return

        if sys.version_info < (3, 8):  # pragma: no cover
            logger.warning(
                "[Gantt] To render weeks, Python 3.8 is the minimum supported version."
            )
            return

        min_date = self._scale_min_date
        max_date = self._scale_max_date

        total_width: float = 0
        weeks_element = etree.SubElement(periods_element, "div", {"class": "weeks"})

        for week_num, week_date in iter_weeks_between_dates(min_date, max_date):
            next_week_date = get_next_week_date(week_num, week_date)
            attributes = {
                "class": f"week week-{week_num}",
                "title": (
                    f"W{week_num}: {week_date.strftime('%Y-%m-%d')} "
                    "&rarr; "
                    f"{(next_week_date - timedelta(days=1)).strftime('%Y-%m-%d')}"
                ),
            }
            element_width: float = 0

            if week_date < self._scale_min_date:
                # the first week must take into account the number of days in the
                # first month of the scale
                element_width = round(
                    self.options.day_width * (next_week_date.day - 1), 2
                )
                attributes.update({"style": f"width: {element_width}px; flex: none;"})
            elif next_week_date > self._scale_max_date:
                # the last week must take into account the number of days in the
                # last month of the scale
                days_diff = (
                    self._scale_max_date
                    - date.fromisocalendar(week_date.year, week_num, 1)
                ).days
                element_width = round(self.options.day_width * (days_diff + 1), 2)
                attributes.update({"style": f"width: {element_width}px; flex: none;"})
            else:
                element_width = self.options.day_width * 7

            week_element = etree.SubElement(weeks_element, "span", attributes)
            week_text_element = etree.SubElement(
                week_element, "span", {"class": "week-text"}
            )
            week_text_element.text = f"W{week_num}"

            # add vertical line
            if self.options.vlines_pace == VerticalLinesPace.WEEKLY:
                total_width += element_width
                context.vertical_lines.append(VerticalLine(round(total_width, 2)))

    def _build_days(self, periods_element):
        min_date = self._scale_min_date
        max_date = self._scale_max_date

        days_element = etree.SubElement(periods_element, "div", {"class": "days"})

        for i, day_date in enumerate(iter_days_between_dates(min_date, max_date)):
            day_element = etree.SubElement(
                days_element,
                "span",
                {
                    "class": f"day day-{i}",
                    "title": day_date.strftime("%Y-%m-%d"),
                },
            )
            day_text_element = etree.SubElement(
                day_element, "span", {"class": "day-text"}
            )
            day_text_element.text = str(day_date.day)

    def _build_groups_html(self, parent, context: BuildContext):
        plan = self.plan
        if plan.activities is None:
            return

        for i, item in enumerate(plan.activities):
            group_element = etree.SubElement(
                parent, "div", {"class": f"nt-plan-group nt-group-{i}"}
            )
            summary_element = etree.SubElement(
                group_element, "div", {"class": "nt-plan-group-summary"}
            )
            p_element = etree.SubElement(summary_element, "p")
            p_element.text = item.title or f"Group {i}"

            activities_element = etree.SubElement(
                group_element, "div", {"class": "nt-plan-group-activities"}
            )

            for line in context.vertical_lines:
                etree.SubElement(
                    activities_element,
                    "i",
                    {"class": "nt-vline", "style": f"left: {line.left}px;"},
                )

            if item.events:
                # TODO: put in common function
                item_element = etree.SubElement(
                    activities_element, "div", {"class": "nt-plan-activity events"}
                )
                actions_element = etree.SubElement(
                    item_element, "div", {"class": "actions"}
                )
                for event in item.events:
                    self.build_event(actions_element, event)

            for activity in item.iter_activities(False):
                self._build_item_html(activities_element, activity)

    def _build_item_html(self, parent, item: Activity):
        item_element = etree.SubElement(parent, "div", {"class": "nt-plan-activity"})
        actions_element = etree.SubElement(item_element, "div", {"class": "actions"})

        title_date = (
            (
                f" {item.start.strftime('%Y-%m-%d')} "
                "&rarr; "
                f"{(item.end + timedelta(days=-1)).strftime('%Y-%m-%d')}"
            )
            if item.start and item.end
            else ""
        )
        period_element = etree.SubElement(
            actions_element,
            "div",
            {
                "class": "period",
                "title": f"{item.title}{title_date}",
                "style": (
                    f"left: {self._calc_period_left(item)}px; "
                    f"width: {self._calc_period_width(item)}px;"
                ),
            },
        )
        span_el = etree.SubElement(period_element, "span")
        span_el.text = item.title

        if item.events:
            for event in item.events:
                self.build_event(actions_element, event)

    def _format_time(self, value: Union[None, date, datetime]):
        if value is None:
            return ""
        if isinstance(value, datetime):
            return value.strftime("%Y-%m-%d, %H:%M:%S")
        return value.strftime("%Y-%m-%d")

    def build_event(self, parent, event: Event):
        dot_element = etree.SubElement(
            parent,
            "div",
            {
                "class": "nt-timeline-dot bigger",
                "title": f"{event.title} {self._format_time(event.time)}",
                "style": f"left: {self._calc_time_left(event.time)-4}px;"
                if event.time
                else "",
            },
        )

        if event.icon:
            build_icon_html(dot_element, event.icon)

    def _calc_time_left(self, time: Union[date, datetime]) -> float:
        delta = (
            time
            - datetime(
                self._scale_min_date.year,
                self._scale_min_date.month,
                self._scale_min_date.day,
                0,
                0,
                0,
            )
            if isinstance(time, datetime)
            else time - self._scale_min_date
        )
        return round(delta.days * self.options.day_width, 2)

    def _calc_period_left(self, activity: Activity) -> float:
        start = activity.start or activity.get_overall_start()
        if start is None:
            return 0

        delta = date_delta(start, self._scale_min_date)
        pixels_width = delta.days * self.options.day_width
        return round(pixels_width, 2)

    def _calc_period_width(self, activity: Activity) -> float:
        start = activity.start or activity.get_overall_start()
        end = activity.end or activity.get_overall_end()
        if end is None or start is None:
            return 50

        delta = date_delta(end, start)
        pixels_width = delta.days * self.options.day_width
        return round(pixels_width, 2)
